// Multiserver by Jeff Breidenbach, copyright May 1996
// Based on Example 9-3 in Java in a Nutshell by David Flanagan
// Flanagan's original copyright is listed below.
//
// Modifications include: 
//   -v verbose switch
//   Multiple channels
//   Star connection between clients
//   Simple handshaking (WELCOME, CHANNEL, OBITUARY) 
//
// Multiserver embodies the philosophy that all intelligence and
// work in a client/server system should be done by the client. The
// server should be as simple as possible, whose only function is
// to relay messages between the clients. Multiserver was designed
// to do as little as possible, and still provide enough power for a
// complex system. 
//
// The complete behavior of Multiserver can be summed up in a
// single paragraph.
//
// Any message sent to Multiserver will be broadcast to all clients
// on the same channel, with the following special cases. 
// On connection, the client will receive the message: 0 WELCOME n
// where n is a unique, non-zero ID number. It will also be assigned
// to a default channel. When a client disconnects, all other clients 
// on the same channel will receive the message: 0 OBITUARY n 
// where n is the ID number of the deceased. Finally, the client
// may change its channel by sending the message: CHANNEL newchannel
// where newchannel is the name of the new channel (string)
// A channel change is not broadcast to anyone.
//
// You may communicate with this server using the Client in example 9-4
// from Java in a Nutshell.
//
// Verion history:
//
// 1.3 Fixed subtle bug in broadcasting, added "quit" capability
// 1.2 Added channel information to the verbose output
// 1.1 Fixed a major bug that was causing channels to bleed into each other
// 1.0 Original release, May 1996

// This example is from the book _Java in a Nutshell_ by David Flanagan.
// Written by David Flanagan.  Copyright (c) 1996 O'Reilly & Associates.
// You may study, use, modify, and distribute this example for any purpose.
// This example is provided WITHOUT WARRANTY either expressed or implied.

import java.net.*;
import java.io.*;
import java.awt.*;
import java.util.*;

public class MultiServer extends Thread {
    protected final static int DEFAULT_PORT = 8411;
    protected int port;
    protected ServerSocket listen_socket;
    protected ThreadGroup threadgroup;
    protected List connection_list;
    protected Vector connections;
    protected Vulture vulture;
    
    // Exit with an error message, when an exception occurs.
    public static void fail(Exception e, String msg) {
        System.err.println(msg + ": " +  e);
        System.exit(1);
    }
    
    // Create a ServerSocket to listen for connections on;  start the thread.
    public MultiServer(int port, boolean verbose) {
        // Create our server thread with a name.
        super("Server");
        if (port == 0) port = DEFAULT_PORT;
        this.port = port;
        try { listen_socket = new ServerSocket(port); }
        catch (IOException e) {fail(e, "Exception creating server socket");}
        // Create a threadgroup for our connections
        threadgroup = new ThreadGroup("Server Connections");

        connection_list = new List();

        // Create a window to display our connections in
        if (verbose) {
           System.out.println("Multiserver version 1.2, connecting to port "
                               + port);
           StatusWindow f = new StatusWindow("MultiServer Status: running");
           f.add("Center", connection_list);
           f.resize(500, 200);
           f.show();
        }

        // Initialize a vector to store our connections in
        connections = new Vector();

        // Create a Vulture thread to wait for other threads to die.
        // It starts itself automatically.
        vulture = new Vulture(this);

        // Start the server listening for connections
        this.start();
    }
    
    // The body of the server thread.  Loop forever, listening for and
    // accepting connections from clients.  For each connection, 
    // create a Connection object to handle communication through the
    // new Socket.  When we create a new connection, add it to the
    // Vector of connections, and display it in the List.  Note that we
    // use synchronized to lock the Vector of connections.  The Vulture
    // class does the same, so the vulture won't be removing dead 
    // connections while we're adding fresh ones.
    public void run() {
        try {
            while(true) {
                Socket client_socket = listen_socket.accept();
                Connection c = new Connection(this, client_socket, threadgroup,
                                  3, vulture);
                // prevent simultaneous access.
                synchronized (connections) {
                    connections.addElement(c);
                    connection_list.addItem(c.toString());
                }
            }
        }
        catch (IOException e) {fail(e, "Exception while listening for connections");}
    }
    
    // Start the server up, listening on an optionally specified port
    public static void main(String[] args) {
        int port = 0;         
        boolean verbose = false;

        if (args.length == 0)  new MultiServer(port, verbose);         
        else if (args.length == 1) {
	    if (args[0].equals("-v"))  {
              verbose = true;
              new MultiServer(port,verbose);
            }
            else {
              try {port = Integer.parseInt(args[0]);} 
              catch (NumberFormatException e) {port = 0;}
              new MultiServer(port, verbose);         
            }
        }

        else if ((args.length == 2) && args[0].equals("-v")) {
            try {port = Integer.parseInt(args[1]);} 
            catch (NumberFormatException e) {port = 0;}
            verbose = true;
            new MultiServer(port, verbose);         
        }

        else System.out.println("Usage: java MultiServer [-v] [<port>]");

    }
}

// This class is the thread that handles all communication with a client
// It also notifies the Vulture when the connection is dropped.
class Connection extends Thread {

    static int connection_number = 0;
    protected Socket client;
    protected Vulture vulture;
    protected DataInputStream in;
    protected PrintStream out;
    protected MultiServer serv;
    protected String channel;

    // Initialize the streams and start the thread
    public Connection(MultiServer theserver, Socket client_socket, ThreadGroup threadgroup, int priority, Vulture vulture) 
    {
        // Give the thread a group, a name, and a priority.
        super(threadgroup, "" + ++connection_number);
        this.setPriority(priority);
        // Save our other arguments away
        client = client_socket;
        this.vulture = vulture;
        serv = theserver;
        channel = "The Default Channel";
        // Create the streams
        try { 
            in = new DataInputStream(client.getInputStream());
            out = new PrintStream(client.getOutputStream());
        }
        catch (IOException e) {
            try {in.close(); out.close(); client.close();} 
            catch (IOException e2) {} ;
            System.err.println("Exception while getting socket streams: " + e);
            return;
        }
        // And start the thread up
        this.start();
    }
    
    // Provide the service.
    // Read a line. If channel command, change channel.
    // Otherwise, broadcast line to all connections on same channel.
    public void run() {
        String line;

        // Send a welcome message to the client
        out.println("0 WELCOME " + super.getName());
          
        try {
            for(;;) {
                // read in a line
                line = in.readLine();
                if (line == null) break;
                 
                if (line.startsWith("CHANNEL ")) {
                  channel = line.substring(8);
                 // This next line is simply for keeping the 
                 // list of connections (for verbose mode)
                 // up to date. I think it may have some serious
                 // performance overhead, and getting rid of this
                 // line means at most there will be a 5 second
                 // lag in channel statistics. Let's get rid of it
                 // and see what happens. 
                 /* synchronized (vulture) {vulture.notify();} */
                }
		else broadcast(line);
            }
        }
        catch (IOException e) {}
        // When we're done, for whatever reason, be sure to close
        // the socket, and to notify the Vulture object.  Note that
        // we have to use synchronized first to lock the vulture
        // object before we can call notify() for it.
        finally {
            try {in.close(); out.close(); client.close();} 
            catch (IOException e2) {};
            synchronized (vulture) {vulture.notify();}
        }
    }

    // This method returns the string representation of the Connection.
    // This is the string that will appear in the GUI List.
    public String toString() {
        return "Connection ID# " + this.getName() + " " 
            + client.getInetAddress().getHostName()
            + " " + channel;
    }


    public void send(String msg) {
        out.println(msg);
        out.flush();
    }
   
    // Here we send a message to everyone who is on the 
    // channel. We'll take a snapshot of all the users on the
    // system, and then send a message to each one on this channel.

    public void broadcast(String msg) {
      Vector snapshot;
      Connection you;

      snapshot = (Vector) serv.connections.clone();
      for (int i=0; i<snapshot.size(); i++) {
              you = (Connection) snapshot.elementAt(i);
	 if (you.channel.equals(channel)) 
	    you.send(msg);
      }  
    }
}

// This class waits to be notified that a thread is dying (exiting)
// and then cleans up the list of threads and the graphical list.
class Vulture extends Thread {
    protected MultiServer server;
    

    protected Vulture(MultiServer s) {
        super(s.threadgroup, "Connection Vulture");
        server = s;
        this.start();
    }

    // This is the method that waits for notification of exiting threads
    // and cleans up the lists.  It is a synchronized method, so it
    // acquires a lock on the `this' object before running.  This is 
    // necessary so that it can call wait() on this.  Even if the 
    // the Connection objects never call notify(), this method wakes up
    // every five seconds and checks all the connections, just in case.
    // Note also that all access to the Vector of connections and to
    // the GUI List component are within a synchronized block as well.
    // This prevents the Server class from adding a new conenction while
    // we're removing an old one.
    public synchronized void run() {
        for(;;) {
            try {this.wait(5000);} catch (InterruptedException e) {};
            // prevent simultaneous access
            synchronized(server.connections) {
                // loop through the connections
                for(int i = 0; i < server.connections.size(); i++) {
                    Connection c;
                    c = (Connection)server.connections.elementAt(i);
                    // if the connection thread isn't alive anymore, 
                    // remove it from the Vector and List. And inform
                    // the deceased's friends.
                    if (!c.isAlive()) {
                        server.connections.removeElementAt(i);
                        server.connection_list.delItem(i);
		        c.broadcast("0 OBITUARY " + c.getName()); 
                        i--;
                    }
                    // if the information for a connection has changed
                    // update the List.
                    else 
                    if (!c.toString().equals(server.connection_list.getItem(i))) {
                        server.connection_list.replaceItem(c.toString(), i);
                    }
                }
            }
        }
    }
}

class StatusWindow extends Frame {

    public StatusWindow(String title)
    {

        super(title); /* shouldn't this be title? */
        this.setForeground(Color.black);
        this.setBackground(Color.lightGray);
    }

    public void go_away() {
            this.hide();	
            this.dispose();
            System.exit(0);
    }	

    public boolean handleEvent(Event event) {
      boolean event_handled = false;
      if (event.id == Event.WINDOW_DESTROY) {
	go_away();
        event_handled = true;
      }	
      if (event_handled) return true;
      else return super.handleEvent(event); 
    }
}





